#include "stdafx.h"
#include <string>
#include <iostream>
#include <memory>
#include <list>

using namespace std;

class Song
{
public:
	Song( const string &title, const string &artist ) :
		Song( title.c_str(), artist.c_str() )							// Delegating Constructor
	{
	}

	Song( const char *pTitle, const char *pArtist ) :
		_title( pTitle ),
		_artist( pArtist )
	{
		cout << "--> Konstruktor von Song.";
		ausgabe();
	}

	Song()
	{
		cout << ">>> Default Konstruktor von Song.";
	}

	Song& operator=(const Song& other)
	{
		cout << ">>> Copy-Operator von Song.";

		if( &other != this )
		{
			_title = other._title;					// ** ich darf nun auf die privates zugreifen!!!
			_artist = other._artist;
		}

		return *this;
	}

public:
	~Song()
	{
		cout << "*** Destruktor von Song.";
		ausgabe();
	}

public:
	const string& getTitle( void ) const
	{
		return _title;
	}

	const string& getArtist( void ) const
	{
		return _artist;
	}

	const void ausgabe( void ) const
	{
		cout << "Title = \"" << _title << "\" von \"" << _artist << "\"\n";
	}

private:
	string _title;
	string _artist;
};

void abspielen( const Song &song )
{
	cout << "abspielen von: ";
	song.ausgabe();
}

int _tmain( int argc, _TCHAR* argv[] )
{
	// ******************************************************************************************************
	cout << "1) der Klassiker\n\n";

	{
		Song *pEinLied = new Song( "Sofa Surfers", "Kruder & Dorfmeister" );
		// cout << "\nK & D didn't left the building...\n";

		// ** und hier habe ich das delete vergessen!!
		// delete pEinLied;
	}

	// ******************************************************************************************************
	cout << "\n\n2) Einfache Memory-Leak-Protection mit uniqui_ptr (entspricht RAII)\n\n";

	{
		unique_ptr<Song> einLied( new Song( "Hip Hip Chin Chin", "Club des Belugas" ) );
		unique_ptr<Song> zweitesLied( new Song( "Tsunami", "Sven van Hees" ) );

		//const string &artist = einLied->getArtist();
		// ....
		cout << endl;
		abspielen( *einLied );
		abspielen( *zweitesLied );
		cout << "\nDie beiden verlassen nun die Buehne.\n\n";
	}

	// ******************************************************************************************************
	cout << "\n\n3) Reference-Counting mit shared_ptr\n\n";

	{
		shared_ptr<Song> unbekannt( nullptr );
		{
			auto song1 = make_shared<Song>( "My Friend", "Groove Armada" );
			unbekannt = song1;

			cout << "\n";
			abspielen( *song1 );
			cout << "\nAnd the gig goes on, and on ...\n\n";
		}

		abspielen( *unbekannt );
		cout << "\nAnd now Grove Armada leaves the building...\n\n";
	}

	// ******************************************************************************************************
	cout << "\n\n4) shared_ptr mit collections\n\n";
	{
		{
			cout << "Und die Kopier-Party beginnt...(kommt bei den rvalues noch genauer)\n\n";
			list<Song> copyPlaylist;
			copyPlaylist.push_back( Song( "Afterworld", "Alex Cortiz" ) );
			copyPlaylist.push_back( Song( "Sunworshipper", "Mylo" ) );
			copyPlaylist.push_back( Song( "Eple", "Roeyksopp" ) );

			cout << "\n";
			for( auto song : copyPlaylist )
			{
				abspielen( song );
			}
			cout << "\nUnd die Kopier-Party ist zu Ende.\n\n";
		}

		{
			cout << "\n\nUnd die Party beginnt...\n\n";
			list<shared_ptr<Song>> playlist;
			playlist.push_back( make_shared<Song>( "Afterworld", "Alex Cortiz" ) );
			playlist.push_back( make_shared<Song>( "Sunworshipper", "Mylo" ) );
			playlist.push_back( make_shared<Song>( "Eple", "Roeyksopp" ) );

			cout << "\n";
			for( auto song : playlist )
			{
				abspielen( *song );
			}
			cout << "\nUnd die Party ist zu Ende.\n\n";
		}
	}

	// ** weak_ptr funktioniert so hnlich wie ein shared_ptr, jedoch ohne reference-counting. Das
	// ** wird gebraucht, wenn man cylce-references abbilden will.

	return 0;
}

// ** In modern C++, raw pointers are only used in small code blocks of limited scope, loops, or helper functions where 
// ** performance is critical and there is no chance of confusion about ownership.