#include "stdafx.h"
#include <string>
#include <iostream>
#include <list>

using namespace std;

class Song
{
public:
	Song( const string &title, const string &artist ) :
		Song( title.c_str(), artist.c_str() )							// Delegating Constructor
	{
	}

	Song( const char *pTitle, const char *pArtist ) :
		_pTitle( nullptr ),
		_pArtist( nullptr )
	{
		cout << "--> Ctor( char*, char* ). ";
		erzeugen( pTitle, pArtist );
	}

	Song( const Song &origin )
	{
		cout << "--> Copy-Ctor. ";
		erzeugen( origin.getTitle().c_str(), origin.getArtist().c_str() );
	}

	Song( Song &&rOrigin )
	{
		cout << "--> Move-Ctor. ";

		_pTitle = rOrigin._pTitle;
		_pArtist = rOrigin._pArtist;

		rOrigin._pTitle = nullptr;
		rOrigin._pArtist = nullptr;

		ausgabe();
	}

public:
	/*Song& operator=(const Song &origin)
	{
		cout << "--> Copy-operator. ";
		if( this != &origin )
		{
			zerstoeren();
			erzeugen( origin.getTitle().c_str(), origin.getArtist().c_str() );
		}
	}*/

private:
	void erzeugen( const char *pTitle, const char *pArtist )
	{
		_pTitle = new string( pTitle );
		_pArtist = new string( pArtist );

		cout << "--> new <-- ";
		ausgabe();
	}

	void zerstoeren( void )
	{
		if( (_pTitle == nullptr) && (_pArtist == nullptr) )
		{
			cout << "+++ es gibt nichts zu zerstoeren +++\n";
		}
		else
		{
			if( _pTitle != nullptr )
			{
				cout << "-> delete <- " << *_pTitle << " ";
				delete _pTitle;
				_pTitle = nullptr;
			}

			if( _pArtist != nullptr )
			{
				cout << "-> delete <- " << *_pArtist << "";
				delete _pArtist;
				_pArtist = nullptr;
			}

			cout << "\n";
		}
	}

public:
	~Song()
	{
		cout << "*** Dtor von Song. ";
		zerstoeren();
	}

public:
	const string& getTitle( void ) const
	{
		return *_pTitle;
	}

	const string& getArtist( void ) const
	{
		return *_pArtist;
	}

	const void ausgabe( void ) const
	{
		cout << "Title = \"" << *_pTitle << "\" von \"" << *_pArtist << "\"\n";
	}

private:
	string *_pTitle;
	string *_pArtist;
};

string gibMirEinenNamen( void )
{
	return "Thomas Kirchmair";
}

void ausgabe( const string &value )
{
	cout << "Ausgabe const-ref: " << value << endl;
}

void ausgabe( string &&value )
{
	cout << "Ausgabe rvalue-ref: " << value << endl;
}

int _tmain( int argc, _TCHAR* argv[] )
{
	// *******************************************************
	// 1) verallgemeinerung (nicht ganz korrekt)
	//	int a = 42;
	//	int b = 43;
	//
	//	// ** a and b are both l-values:
	//	a = b;					// ok
	//	b = a;					// ok
	//	a = a * b;				// ok
	//
	//	//**  a * b is an rvalue:
	//	int c = a * b;			// ok, rvalue on right hand side of assignment
	//	a * b = 42;				// error, rvalue on left hand side of assignment

	// *******************************************************
	// 2) genauere Defintion
	// lvalues:
	//int i = 42;
	//i = 43;				// ok, i is an lvalue
	//int* p = &i;			// ok, i is an lvalue
	//int& foo( );
	//foo( ) = 42;			// ok, foo() is an lvalue
	//int* p1 = &foo( );	// ok, foo() is an lvalue

	//// rvalues:
	//int foobar( );
	//int j = 0;
	//j = foobar( );			// ok, foobar() is an rvalue
	//int* p2 = &foobar( );		// error, cannot take the address of an rvalue
	//j = 42;					// ok, 42 is an rvalue

	// *******************************************************
	// --> and rvalue-reference is a reference that will only bind to a temporary object.
	// --> hiermit kann unterschieden werden, ob auf ein temporres Objekt (rvalue), oder auf ein
	// Objekt mit einem Namen zugegriffen wird.
	// ** can't use a mutuable reference.
	string ergebnis = gibMirEinenNamen();					// ok, aber da wird viel kopiert.
	//	string &ergebnis = gibMirEinenNamen( );					// geht nicht, weil temporr
	const string& refName = gibMirEinenNamen();			// ok, aber const
	string &&rtemp = gibMirEinenNamen();

	string test( "Tom" );
	ausgabe( test );
	ausgabe( gibMirEinenNamen() );

	// *******************************************************
	// 3) move-Semantik (Copy-Problem auf mehreren Ebenen
	// Lsung 1) Copy-Konstruktor schreiben, weil der Default die Pointer kopiert
	{
		cout << "Und die Kopier-Party beginnt...\n\n";
		list<Song> copyPlaylist;
		copyPlaylist.push_back( Song( "Afterworld", "Alex Cortiz" ) );
		copyPlaylist.push_back( Song( "Sunworshipper", "Mylo" ) );
		copyPlaylist.push_back( Song( "Eple", "Roeyksopp" ) );

		cout << "\n";
		for( auto &song : copyPlaylist )
		{
			song.ausgabe();   // <- Exception fr den Default-Fall
		}

		cout << "\nUnd die Kopier-Party ist zu Ende.\n\n";
	}

	return 0;
}

// http://thbecker.net/articles/rvalue_references/section_01.html
