#include "stdafx.h"
#include <iostream>
#include <initializer_list>
#include <stdarg.h>

using namespace std;

// ******************************************************************************************************
// ** Output function to output any type without type specifiers like printf() family
template <typename _Type, typename ..._VariadicParams> void output( _Type value, _VariadicParams ...params )
{
	cout << value << ' ';
	if( sizeof...(params) > 0 )
	{
		output( params... );
	}
	else
	{
		cout << std::endl;
	}
}

// ** Since variadic templates are recursive, must have a base case
void output( void )
{
	cout << std::endl;
}

template <typename _Type> _Type sum( const _Type &value )
{
	return value;
}

template <typename _Type, typename ..._VariadicParameters> _Type sum( _Type value, _VariadicParameters ...params )
{
	if( sizeof...(params) > 0 )
	{
		value += sum( params... );
	}

	return value;
}


// ******************************************************************************************************
#define AUSGABE( format, ... ) printf( format, __VA_ARGS__ )
#define CALL_TESTFUNCTION( ... ) testFunction( __VA_ARGS__ )

void testFunction( int value1, int value2, int value3, int value4 )
{
	cout << value1 << " " << value2 << " " << value3 << " " << value4 << std::endl;
}


// ******************************************************************************************************
template <class _Type> void initFunction( std::initializer_list<_Type> initList )
{
	for( auto elem : initList )
	{
		std::cout << elem << std::endl;
	}
}


// ******************************************************************************************************
int maxof( int n_args, ... )
{
	va_list vaList;
	va_start( vaList, n_args );
	int max = va_arg( vaList, int );
	for( int i = 2; i <= n_args; i++ )
	{
		int a = va_arg( vaList, int );
		if( a > max )
		{
			max = a;
		}
	}
	va_end( vaList );

	return max;
}

int _tmain(int argc, _TCHAR* argv[])
{
	// ***********************************************************
	// ** 1) Ellipsis Operator
	try
	{
	}
	catch( ... )
	{
	}

	// ***********************************************************
	// ** 2) Ellipsis und define (var_args)
	AUSGABE( "%s, %s, %d, %f\n", "string1", "wert2", 24, 23.33 );
	CALL_TESTFUNCTION( 12, 34, 56, 78 );


	// ***********************************************************
	// ** 3) Function with initialiser list
	initFunction( { 10, 20, 30, 40 } ) ;


	// ***********************************************************
	// ** 4) "old" va_arg-Functions
	int result = maxof( 5, 99, 111, 22, 44, 55 );


	// ***********************************************************
	// ** 5) Variadic Templates
	output( );
	output( '5' );
	output( '5', 2 );
	output( '5', 2, "muh" );
	output( '5', 2, "muh", -1 );
	output( '5', 2, "muh", -1, 0.5f );
	output( '5', 2, "muh", -1, 0.5f, 16.3 );

	cout << endl;

	cout << sum( 1 ) << '\n';
	cout << sum( 1, 2 ) << '\n';
	cout << sum( 1, 2, 3 ) << '\n';
	cout << sum( 1, 2, 3, 4 ) << '\n';
	cout << sum( 1.23, 2.0, 3.34, 4.222, 5 ) << '\n';

	return 0;
}

